﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using ESRI.ArcGIS.Client;
using ESRI.ArcGIS.Client.Geometry;

namespace $safeprojectname$
{
    public partial class MeasureDialog : UserControl
    {

        private enum GeomType
        {
            Point = 0,
            Line = 1,
            Poly = 2
        }

        // Currently supported map units are decimal degrees, meters, and feet
        // Supported length and area units are as below
        // (You can easily add support for other map and measurement units)
        private string[] _LengthUnitLabels = new string[] { "Miles", "Feet", "Meters", "km" };
        private string[] _AreaUnitLabels = new string[] { "Sq Mi", "Acres", "Sq Ft", "Sq m", "Sq km" };

        private MainPage _Parent;
        private Map _Map;
        private ScaleBarUnit _MapUnit;
        private bool _bGeodesic = false;
        private double _CF; // Conversion factor map units to meters
        private string _sCF; // Decimal places for coordinate display
        private string _GraphicsLayer = "";

        private Button _btnCurrentTool = null;
        private GeomType _GeomType;
        private Draw _MeasureDrawObject;
        private bool _bDrawBegun;

        private bool _bAddNew;
        private MapPoint _ptCurrent;
        private ESRI.ArcGIS.Client.Geometry.PointCollection _MeasurePoints = new ESRI.ArcGIS.Client.Geometry.PointCollection();
        private double _dUpperResult = 0;
        private double _dLowerResult = 0;

        #region Public Methods
        public MeasureDialog()
        {
            InitializeComponent();
            MyDialogBase.SetParent(this);
            MyDialogBase.CloseEvent += MyDialogBase_Close;
            MyDialogBase.CollapseEvent += MyDialogBase_Collapse;
            MyDialogBase.ExpandEvent += MyDialogBase_Expand;
            btnPoint.Click += btnPoint_Click;
            btnLine.Click += btnLine_Click;
            btnPoly.Click += btnPoly_Click;
            cbUpper.SelectionChanged += cbUpper_SelectionChanged;
            cbLower.SelectionChanged += cbLower_SelectionChanged;
        }

        public void SetParent(MainPage TheParent)
        {
            // Assumes that MainPage supports SetCursor and UnsetTool
            _Parent = TheParent;
            lblUpperText.Text = "";
            cbUpper.Visibility = Visibility.Collapsed;
            lblLowerText.Text = "";
            cbLower.Visibility = Visibility.Collapsed;
        }

        public void SetMap(Map TheMap)
        {
            _Map = TheMap;
            _Map.MouseClick += _Map_MouseClick;
            _Map.MouseMove += _Map_MouseMove;
            _MeasureDrawObject = new Draw(_Map) { FillSymbol = DefaultFillSymbol, LineSymbol = DefaultLineSymbol };
            _MeasureDrawObject.DrawBegin += _MeasureDrawObject_DrawBegin;
            _MeasureDrawObject.VertexAdded += _MeasureDrawObject_VertexAdded;
            _MeasureDrawObject.DrawComplete += _MeasureDrawObject_DrawComplete;
        }

        public bool SetMapUnits(ScaleBarUnit TheUnits)
        {
            _MapUnit = TheUnits;
            _CF = 1.0;
            _sCF = "0.##";
            _bGeodesic = false;
            switch (TheUnits)
            {
                case ScaleBarUnit.DecimalDegrees:
                    _sCF = "0.######";
                    _bGeodesic = true;
                    break;
                case ScaleBarUnit.Meters:
                    break;
                case ScaleBarUnit.Feet:
                    _CF = 0.3048;
                    break;
                default:
                    return false;
            }
            if (_bGeodesic)
                InitGeoAreaParams();
            return true;
        }

        public void SetGraphicsLayer(string sLayer)
        {
            _GraphicsLayer = sLayer;
        }

        public void UnsetTool()
        {
            if (_btnCurrentTool == null)
                return;
            _btnCurrentTool.Background = (Brush) Application.Current.Resources["ButtonNormalBrush"];
            _MeasureDrawObject.IsEnabled = false;
            _btnCurrentTool = null;
        }
        #endregion

        #region Handlers
        private void MyDialogBase_Close(Object sender, EventArgs e)
        {
            GraphicsLayer graphicsLayer = (GraphicsLayer)_Map.Layers[_GraphicsLayer];
            graphicsLayer.ClearGraphics();
            _MeasurePoints.Clear();
            this.Visibility = Visibility.Collapsed;
            if (_btnCurrentTool != null)
                _Parent.UnsetTool();
        }

        private void MyDialogBase_Collapse(Object sender, EventArgs e)
        {
            MeasurePanel.Visibility = Visibility.Collapsed;
        }

        private void MyDialogBase_Expand(Object sender, EventArgs e)
        {
            MeasurePanel.Visibility = Visibility.Visible;
        }

        private void btnPoint_Click(Object sender, RoutedEventArgs e)
        {
            SetTool(btnPoint);
        }

        private void btnLine_Click(Object sender, RoutedEventArgs e)
        {
            SetTool(btnLine);
        }

        private void btnPoly_Click(Object sender, RoutedEventArgs e)
        {
            SetTool(btnPoly);
        }

        private void cbUpper_SelectionChanged(Object sender, SelectionChangedEventArgs e)
        {
            UpdateValues();
        }

        private void cbLower_SelectionChanged(Object sender, SelectionChangedEventArgs e)
        {
            UpdateValues();
        }

        private void _Map_MouseClick(Object sender, Map.MouseEventArgs e)
        {
            if (_btnCurrentTool == null)
                return;
            else if (_btnCurrentTool != btnPoint)
                return;
            GraphicsLayer graphicsLayer = (GraphicsLayer)_Map.Layers[_GraphicsLayer];
            graphicsLayer.ClearGraphics();
            Graphic graphic = new Graphic() { Geometry = _ptCurrent, Symbol = DefaultMarkerSymbol };
            graphicsLayer.Graphics.Add(graphic);
            _bDrawBegun = false;
            _Parent.UnsetTool();
        }

        private void _Map_MouseMove(Object sender, MouseEventArgs args)
        {
            if (_Map.Extent == null)
                return;
            if (_btnCurrentTool == null)
                return;
            Point screenPt = args.GetPosition(_Map);
            _ptCurrent = _Map.ScreenToMap(screenPt);
            if (!_bDrawBegun)
                return;
            if (_btnCurrentTool == btnPoint)
            {
                lblUpperText.Text = "X: " + _ptCurrent.X.ToString(_sCF);
                lblLowerText.Text = "Y: " + _ptCurrent.Y.ToString(_sCF);
                return;
            }
            if (_bAddNew)
            {
                AddCurrentPoint();
                _bAddNew = false;
            }
            else
            {
                int iCount = _MeasurePoints.Count;
                if (iCount == 0)
                    return;
                _MeasurePoints[iCount - 1].X = _ptCurrent.X;
                _MeasurePoints[iCount - 1].Y = _ptCurrent.Y;
            }
            UpdateGeometry();
        }

        private void _MeasureDrawObject_DrawBegin(Object sender, EventArgs args)
        {
            if (_bDrawBegun)
                return;
            _bDrawBegun = true;
            _dUpperResult = 0;
            _dLowerResult = 0;
            _MeasurePoints.Clear();
            AddCurrentPoint();
            UpdateGeometry();
            _bAddNew = true;
        }

        private void _MeasureDrawObject_VertexAdded(Object sender, VertexAddedEventArgs args)
        {
            _bAddNew = true;
        }

        private void _MeasureDrawObject_DrawComplete(Object sender, DrawEventArgs args)
        {
            GraphicsLayer graphicsLayer = (GraphicsLayer)_Map.Layers[_GraphicsLayer];
            graphicsLayer.ClearGraphics();
            ESRI.ArcGIS.Client.Symbols.Symbol sym;
            if (_GeomType == GeomType.Line)
                sym = DefaultLineSymbol;
            else
                sym = DefaultFillSymbol;
            Graphic graphic = new Graphic() { Geometry = args.Geometry, Symbol = sym };
            graphicsLayer.Graphics.Add(graphic);
            _bDrawBegun = false;
        }
        #endregion

        #region Private Methods
        private void SetTool(Button btn)
        {
            _Parent.UnsetTool();
            btn.Background = (Brush) Application.Current.Resources["ButtonHighlightBrush"];
            _btnCurrentTool = btn;
            _dUpperResult = 0;
            _dLowerResult = 0;
            _MeasurePoints.Clear();
            _bDrawBegun = false;
            _bAddNew = false;
            if (btn == btnPoint)
            {
                lblUpperText.Text = "X: ";
                cbUpper.Visibility = Visibility.Collapsed;
                lblLowerText.Text = "Y: ";
                cbLower.Visibility = Visibility.Collapsed;
                _GeomType = GeomType.Point;
                _MeasureDrawObject.IsEnabled = false;
                _bDrawBegun = true;
            }
            else if (btn == btnLine)
            {
                lblUpperText.Text = "Segment: ";
                cbUpper.Visibility = Visibility.Collapsed;
                lblLowerText.Text = "Total: ";
                cbLower.Items.Clear();
                foreach (string sItem in _LengthUnitLabels)
                    cbLower.Items.Add(sItem);
                cbLower.Visibility = Visibility.Visible;
                cbLower.SelectedIndex = 0;
                _GeomType = GeomType.Line;
                _MeasureDrawObject.DrawMode = DrawMode.Polyline;
                _MeasureDrawObject.IsEnabled = true;
            }
            else
            {
                lblUpperText.Text = "Perimeter: ";
                cbUpper.Items.Clear();
                foreach (string sItem in _LengthUnitLabels)
                    cbUpper.Items.Add(sItem);
                cbUpper.Visibility = Visibility.Visible;
                cbUpper.SelectedIndex = 0;
                lblLowerText.Text = "Area: ";
                cbLower.Items.Clear();
                foreach (string sItem in _AreaUnitLabels)
                    cbLower.Items.Add(sItem);
                cbLower.Visibility = Visibility.Visible;
                cbLower.SelectedIndex = 0;
                _GeomType = GeomType.Poly;
                _MeasureDrawObject.DrawMode = DrawMode.Polygon;
                _MeasureDrawObject.IsEnabled = true;
            }
            _Parent.SetCursor(MouseCursorType.Measure);
        }

        private void AddCurrentPoint()
        {
            MapPoint pt = new MapPoint(_ptCurrent.X, _ptCurrent.Y);
            pt.SpatialReference = _Map.SpatialReference;
            _MeasurePoints.Add(pt);
        }

        private void UpdateGeometry()
        {

            int iCount;
            double dLen, dSegLen, dArea;

            iCount = _MeasurePoints.Count();
            if (iCount < 2)
                return;
            if (_GeomType == GeomType.Line)
            {
                if (_bGeodesic)
                {
                    dLen = GeoLength(false);
                    dSegLen = GeoDist(_MeasurePoints[iCount - 2], _MeasurePoints[iCount - 1]);
                }
                else
                {
                    dLen = PlanarLength(false);
                    dSegLen = Dist(_MeasurePoints[iCount - 2], _MeasurePoints[iCount - 1]);
                }
                _dUpperResult = dSegLen * _CF;
                _dLowerResult = dLen * _CF;
                UpdateValues();
                return;
            }
            if (iCount < 3)
            {
                if (_bGeodesic)
                    dLen = GeoLength(false);
                else
                    dLen = PlanarLength(false);
                _dUpperResult = dLen * _CF;
                _dLowerResult = 0;
                UpdateValues();
                return;
            }
            if (_bGeodesic)
            {
                dLen = GeoLength(true);
                dArea = GeoArea();
            }
            else
            {
                dLen = PlanarLength(true);
                dArea = PlanarArea();
            }
            _dUpperResult = dLen * _CF;
            _dLowerResult = dArea * _CF * _CF;
            UpdateValues();
        }

        private void UpdateValues()
        {
            if ((_dUpperResult == 0) && (_dLowerResult == 0))
                return;
            if (_GeomType == GeomType.Line)
            {
                lblUpperText.Text = "Segment: " + ConvertLength(_dUpperResult);
                lblLowerText.Text = "Total: " + ConvertLength(_dLowerResult);
            }
            else
            {
                lblUpperText.Text = "Perimeter: " + ConvertLength(_dUpperResult);
                lblLowerText.Text = "Area: " + ConvertArea(_dLowerResult);
            }
        }

        private string ConvertLength(double dMeters)
        {
            int i;
            if (_btnCurrentTool == btnLine)
                i = cbLower.SelectedIndex;
            else
                i = cbUpper.SelectedIndex;
            double dVal = 0;
            switch (i)
            {
                case 0: // Miles
                    dVal = dMeters / 0.3048 / 5280;
                    break;
                case 1: // Feet
                    dVal = dMeters / 0.3048;
                    break;
                case 2: // Meters
                    dVal = dMeters;
                    break;
                case 3: // Kilometers
                    dVal = dMeters / 1000;
                    break;
            }
            return dVal.ToString("0.##");
        }

        private string ConvertArea(double dSqMeters)
        {
            int i = cbLower.SelectedIndex;
            double dVal = 0;
            switch (i)
            {
                case 0: // Sq. Miles
                    dVal = dSqMeters / 0.09290304 / 27878400;
                    break;
                case 1: // Acres
                    dVal = dSqMeters / 0.09290304 / 43560;
                    break;
                case 2: // Sq. Feet
                    dVal = dSqMeters / 0.09290304;
                    break;
                case 3: // Sq. Meters
                    dVal = dSqMeters;
                    break;
                case 4: // Sq. Kilometers
                    dVal = dSqMeters / 1000000;
                    break;
            }
            return dVal.ToString("0.####");
        }

        private double PlanarLength(bool bClose)
        {
            int iCount, i;
            double dLen = 0;
            iCount = _MeasurePoints.Count;
            if (iCount < 2)
                return dLen;
            for (i = 0; i < iCount - 1; i++)
                dLen += Dist(_MeasurePoints[i], _MeasurePoints[i + 1]);
            if (bClose)
                dLen += Dist(_MeasurePoints[iCount - 1], _MeasurePoints[0]);
            return dLen;
        }

        private double PlanarArea()
        {
            double dArea = 0;
            int iCount, i, j;
            iCount = _MeasurePoints.Count;
            if (iCount < 3)
                return 0;
            for (i = 0; i < iCount; i++)
            {
                if (i == iCount - 1)
                    j = 0;
                else
                    j = i + 1;
                dArea += Determinant(_MeasurePoints[i], _MeasurePoints[j]);
            }
            return Math.Abs(dArea / 2);
        }

        private double Dist(MapPoint pt1, MapPoint pt2)
        {
            double d1, d2;
            d1 = pt2.X - pt1.X;
            d2 = pt2.Y - pt1.Y;
            return Math.Sqrt((d1 * d1) + (d2 * d2));
        }

        private double Determinant(MapPoint pt1, MapPoint pt2)
        {
            return (pt1.X * pt2.Y) - (pt2.X * pt1.Y);
        }
        #endregion

        #region Geodesic Measurement Functions 
        //**** From GRASS source code ****

        // WGS-84 ellipsoid params
        private const double _a = 6378137;
        private const double _b = 6356752.314;

        // Geodesic distance

        private double GeoLength(bool bClose)
        {

            int iCount, i;
            double dLen = 0;

            iCount = _MeasurePoints.Count;
            if (iCount < 2)
                return dLen;
            for (i = 0; i < iCount - 1; i++)
                dLen += GeoDist(_MeasurePoints[i], _MeasurePoints[i + 1]);
            if (bClose)
                dLen += GeoDist(_MeasurePoints[iCount - 1], _MeasurePoints[0]);
            return dLen;

        }

        private double GeoDist(MapPoint pt1, MapPoint pt2)
        {

            double lon1 = pt1.X * Math.PI / 180;
            double lat1 = pt1.Y * Math.PI / 180;
            double lon2 = pt2.X * Math.PI / 180;
            double lat2 = pt2.Y * Math.PI / 180;
            double e2 = (_a * _a - _b * _b) / (_a * _a);

            double boa, f, ff64, al, t1, t2, t3, t4, t1r, t2r;
            double stm, ctm, sdtm, cdtm, tm, dtm;

            al = _a;
            boa = Math.Sqrt(1 - e2);
            f = 1 - boa;
            ff64 = f * f / 64;

            t1r = Math.Atan(boa * Math.Tan(lat1));
            t2r = Math.Atan(boa * Math.Tan(lat2));
            tm = (t1r + t2r) / 2;
            dtm = (t2r - t1r) / 2;
            stm = Math.Sin(tm);
            ctm = Math.Cos(tm);
            sdtm = Math.Sin(dtm);
            cdtm = Math.Cos(dtm);
            t1 = stm * cdtm;
            t1 = t1 * t1 * 2;
            t2 = sdtm * ctm;
            t2 = t2 * t2 * 2;
            t3 = sdtm * sdtm;
            t4 = cdtm * cdtm - stm * stm;

            double a, cd, d, e, q, sd, sdlmr, t, u, v, x, y;

            sdlmr = Math.Sin((lon2 - lon1) / 2);
            if ((sdlmr == 0) && (t1r == t2r))
                return 0;
            q = t3 + sdlmr * sdlmr * t4;
            if (q == 1)
                return Math.PI * al;
            cd = 1 - 2 * q;
            sd = 2 * Math.Sqrt(q - q * q);
            if ((q != 0) && (cd == 1))
                t = 1;
            else if (sd == 0)
                t = 1;
            else
                t = Math.Acos(cd) / sd;

            u = t1 / (1 - q);
            v = t2 / q;
            d = 4 * t * t;
            x = u + v;
            e = -2 * cd;
            y = u - v;
            a = -d * e;

            return (al * sd * (t - f / 4 * (t * x - y) + ff64 *
                    (x * (a + (t - (a + e) / 2) * x) + y * (-2 * d + e * y) + d * x * y)));
        
        }

        // Geodesic area

        private double QA, QB, QC;
        private double QbarA, QbarB, QbarC, QbarD;
        private double AE, Qp, E;

        private void InitGeoAreaParams()
        {

            double e2, e4, e6;

            e2 = (_a * _a - _b * _b) / (_a * _a);
            e4 = e2 * e2;
            e6 = e4 * e2;

            AE = _a * _a * (1 - e2);
            QA = (2 / 3) * e2;
            QB = (3 / 5) * e4;
            QC = (4 / 7) * e6;
            QbarA = -1 - (2 / 3) * e2 - (3 / 5) * e4 - (4 / 7) * e6;
            QbarB = (2 / 9) * e2 + (2 / 5) * e4 + (4 / 7) * e6;
            QbarC = -(3 / 25) * e4 - (12 / 35) * e6;
            QbarD = (4 / 49) * e6;

            Qp = Q(Math.PI / 2);
            E = 4 * Math.PI * Qp * AE;
            if (E < 0)
                E = -E;

        }

        private double GeoArea()
        {

            int iCount, i;
            double x1, y1, x2, y2, dx, dy, Qbar1, Qbar2, area;

            iCount = _MeasurePoints.Count;
            if (iCount < 3)
                return 0;
            x2 = _MeasurePoints[iCount - 1].X * Math.PI / 180;
            y2 = _MeasurePoints[iCount - 1].Y * Math.PI / 180;
            Qbar2 = Qbar(y2);
            area = 0;
            for (i = 0; i < iCount; i++)
            {
                x1 = x2;
                y1 = y2;
                Qbar1 = Qbar2;
                x2 = _MeasurePoints[i].X * Math.PI / 180;
                y2 = _MeasurePoints[i].Y * Math.PI / 180;
                Qbar2 = Qbar(y2);
                if (x1 > x2)
                {
                    while (x1 - x2 > Math.PI)
                        x2 += Math.PI + Math.PI;
                }
                else if (x2 > x1)
                {
                    while (x2 - x1 > Math.PI)
                        x1 += Math.PI + Math.PI;
                }
                dx = x2 - x1;
                area += dx * (Qp - Q(y2));
                dy = y2 - y1;
                if (dy != 0)
                    area += dx * Q(y2) - (dx / dy) * (Qbar2 - Qbar1);
            }
            area *= AE;
            if (area < 0)
                area = -area;
            if (area > E)
                area = E;
            if (area > E / 2)
                area = E - area;
            return area;

        }

        private double Q(double x)
        {
            double sinx, sinx2;
            sinx = Math.Sin(x);
            sinx2 = sinx * sinx;
            return sinx * (1 + sinx2 * (QA + sinx2 * (QB + sinx2 * QC)));
        }

        private double Qbar(double x)
        {
            double cosx, cosx2;
            cosx = Math.Cos(x);
            cosx2 = cosx * cosx;
            return cosx * (QbarA + cosx2 * (QbarB + cosx2 * (QbarC + cosx2 * QbarD)));
        }
        #endregion

    }
}
