﻿Imports ESRI.ArcGIS.Client
Imports ESRI.ArcGIS.Client.Geometry

Partial Public Class MeasureDialog
    Inherits UserControl

    Private Enum GeomType
        Point = 0
        Line = 1
        Poly = 2
    End Enum

    ' Currently supported map units are decimal degrees, meters, and feet
    ' Supported length and area units are as below
    ' (You can easily add support for other map and measurement units)
    Private _LengthUnitLabels() As String = New String() {"Miles", "Feet", "Meters", "km"}
    Private _AreaUnitLabels() As String = New String() {"Sq Mi", "Acres", "Sq Ft", "Sq m", "Sq km"}

    Private _Parent As MainPage
    Private _Map As Map
    Private _MapUnit As ScaleBarUnit
    Private _bGeodesic As Boolean = False
    Private _CF As Double ' Conversion factor map units to meters
    Private _sCF As String ' Decimal places for coordinate display
    Private _GraphicsLayer As String = ""

    Private _btnCurrentTool As Button = Nothing
    Private _GeomType As GeomType
    Private _MeasureDrawObject As Draw
    Private _bDrawBegun As Boolean

    Private _bAddNew As Boolean
    Private _ptCurrent As MapPoint
    Private _MeasurePoints As New PointCollection
    Private _dUpperResult As Double = 0
    Private _dLowerResult As Double = 0

#Region "Public methods"
    Public Sub New()
        InitializeComponent()
        MyDialogBase.SetParent(Me)
        AddHandler MyDialogBase.CloseEvent, AddressOf MyDialogBase_Close
        AddHandler MyDialogBase.CollapseEvent, AddressOf MyDialogBase_Collapse
        AddHandler MyDialogBase.ExpandEvent, AddressOf MyDialogBase_Expand
    End Sub

    Public Sub SetParent(ByVal TheParent As MainPage)
        ' Assumes that MainPage supports SetCursor and UnsetTool
        _Parent = TheParent
        lblUpperText.Text = ""
        cbUpper.Visibility = Visibility.Collapsed
        lblLowerText.Text = ""
        cbLower.Visibility = Visibility.Collapsed
    End Sub

    Public Sub SetMap(ByVal TheMap As Map)
        _Map = TheMap
        AddHandler _Map.MouseClick, AddressOf _Map_MouseClick
        AddHandler _Map.MouseMove, AddressOf _Map_MouseMove
        _MeasureDrawObject = New Draw(_Map) With {.FillSymbol = DefaultFillSymbol, .LineSymbol = DefaultLineSymbol}
        AddHandler _MeasureDrawObject.DrawBegin, AddressOf _MeasureDrawObject_DrawBegin
        AddHandler _MeasureDrawObject.VertexAdded, AddressOf _MeasureDrawObject_VertexAdded
        AddHandler _MeasureDrawObject.DrawComplete, AddressOf _MeasureDrawObject_DrawComplete
    End Sub

    Public Function SetMapUnits(ByVal TheUnits As ScaleBarUnit) As Boolean
        _MapUnit = TheUnits
        _CF = 1.0
        _sCF = "0.##"
        _bGeodesic = False
        Select Case TheUnits
            Case ScaleBarUnit.DecimalDegrees
                _sCF = "0.######"
                _bGeodesic = True
            Case ScaleBarUnit.Meters
                Exit Select
            Case ScaleBarUnit.Feet
                _CF = 0.3048
            Case Else
                Return False
        End Select
        If _bGeodesic Then
            InitGeoAreaParams()
        End If
        Return True
    End Function

    Public Sub SetGraphicsLayer(ByVal sLayer As String)
        _GraphicsLayer = sLayer
    End Sub

    Public Sub UnsetTool()
        If _btnCurrentTool Is Nothing Then
            Exit Sub
        End If
        _btnCurrentTool.Background = Application.Current.Resources("ButtonNormalBrush")
        _MeasureDrawObject.IsEnabled = False
        _btnCurrentTool = Nothing
    End Sub
#End Region

#Region "Handlers"
    Private Sub MyDialogBase_Close(ByVal sender As Object, ByVal e As EventArgs)
        Dim graphicsLayer As GraphicsLayer = TryCast(_Map.Layers(_GraphicsLayer), GraphicsLayer)
        graphicsLayer.ClearGraphics()
        _MeasurePoints.Clear()
        Me.Visibility = Visibility.Collapsed
        If Not _btnCurrentTool Is Nothing Then
            _Parent.UnsetTool()
        End If
    End Sub

    Private Sub MyDialogBase_Collapse(ByVal sender As Object, ByVal e As EventArgs)
        MeasurePanel.Visibility = Visibility.Collapsed
    End Sub

    Private Sub MyDialogBase_Expand(ByVal sender As Object, ByVal e As EventArgs)
        MeasurePanel.Visibility = Visibility.Visible
    End Sub

    Private Sub btnPoint_Click(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles btnPoint.Click
        SetTool(btnPoint)
    End Sub

    Private Sub btnLine_Click(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles btnLine.Click
        SetTool(btnLine)
    End Sub

    Private Sub btnPoly_Click(ByVal sender As System.Object, ByVal e As System.Windows.RoutedEventArgs) Handles btnPoly.Click
        SetTool(btnPoly)
    End Sub

    Private Sub cbUpper_SelectionChanged(ByVal sender As System.Object, ByVal e As System.Windows.Controls.SelectionChangedEventArgs) Handles cbUpper.SelectionChanged
        UpdateValues()
    End Sub

    Private Sub cbLower_SelectionChanged(ByVal sender As System.Object, ByVal e As System.Windows.Controls.SelectionChangedEventArgs) Handles cbLower.SelectionChanged
        UpdateValues()
    End Sub

    Private Sub _Map_MouseClick(ByVal sender As Object, ByVal e As ESRI.ArcGIS.Client.Map.MouseEventArgs)
        If _btnCurrentTool Is Nothing Then
            Exit Sub
        ElseIf Not _btnCurrentTool Is btnPoint Then
            Exit Sub
        End If
        Dim graphicsLayer As GraphicsLayer = TryCast(_Map.Layers(_GraphicsLayer), GraphicsLayer)
        graphicsLayer.ClearGraphics()
        Dim graphic As New ESRI.ArcGIS.Client.Graphic() With {.Geometry = _ptCurrent, .Symbol = DefaultMarkerSymbol}
        graphicsLayer.Graphics.Add(graphic)
        _bDrawBegun = False
        _Parent.UnsetTool()
    End Sub

    Private Sub _Map_MouseMove(ByVal sender As Object, ByVal args As System.Windows.Input.MouseEventArgs)
        If _Map.Extent Is Nothing Then
            Exit Sub
        End If
        If _btnCurrentTool Is Nothing Then
            Exit Sub
        End If
        Dim screenPt As System.Windows.Point = args.GetPosition(_Map)
        _ptCurrent = _Map.ScreenToMap(screenPt)
        If Not _bDrawBegun Then
            Exit Sub
        End If
        If _btnCurrentTool Is btnPoint Then
            lblUpperText.Text = "X: " & _ptCurrent.X.ToString(_sCF)
            lblLowerText.Text = "Y: " & _ptCurrent.Y.ToString(_sCF)
            Exit Sub
        End If
        If _bAddNew Then
            AddCurrentPoint()
            _bAddNew = False
        Else
            Dim iCount As Integer = _MeasurePoints.Count
            If iCount = 0 Then
                Exit Sub
            End If
            _MeasurePoints(iCount - 1).X = _ptCurrent.X
            _MeasurePoints(iCount - 1).Y = _ptCurrent.Y
        End If
        UpdateGeometry()
    End Sub

    Private Sub _MeasureDrawObject_DrawBegin(ByVal sender As Object, ByVal args As EventArgs)
        If _bDrawBegun Then
            Exit Sub
        End If
        _bDrawBegun = True
        _dUpperResult = 0
        _dLowerResult = 0
        _MeasurePoints.Clear()
        AddCurrentPoint()
        UpdateGeometry()
        _bAddNew = True
    End Sub

    Private Sub _MeasureDrawObject_VertexAdded(ByVal sender As Object, ByVal args As VertexAddedEventArgs)
        _bAddNew = True
    End Sub

    Private Sub _MeasureDrawObject_DrawComplete(ByVal sender As Object, ByVal args As DrawEventArgs)
        Dim graphicsLayer As GraphicsLayer = TryCast(_Map.Layers(_GraphicsLayer), GraphicsLayer)
        graphicsLayer.ClearGraphics()
        Dim sym As Symbols.Symbol
        If _GeomType = GeomType.Line Then
            sym = DefaultLineSymbol
        Else
            sym = DefaultFillSymbol
        End If
        Dim graphic As New ESRI.ArcGIS.Client.Graphic() With {.Geometry = args.Geometry, .Symbol = sym}
        graphicsLayer.Graphics.Add(graphic)
        _bDrawBegun = False
    End Sub
#End Region

#Region "Private methods"
    Private Sub SetTool(ByVal btn As Button)
        _Parent.UnsetTool()
        btn.Background = Application.Current.Resources("ButtonHighlightBrush")
        _btnCurrentTool = btn
        _dUpperResult = 0
        _dLowerResult = 0
        _MeasurePoints.Clear()
        _bDrawBegun = False
        _bAddNew = False
        If btn Is btnPoint Then
            lblUpperText.Text = "X: "
            cbUpper.Visibility = Visibility.Collapsed
            lblLowerText.Text = "Y: "
            cbLower.Visibility = Visibility.Collapsed
            _GeomType = GeomType.Point
            _MeasureDrawObject.IsEnabled = False
            _bDrawBegun = True
        ElseIf btn Is btnLine Then
            lblUpperText.Text = "Segment: "
            cbUpper.Visibility = Visibility.Collapsed
            lblLowerText.Text = "Total: "
            cbLower.Items.Clear()
            For Each sItem As String In _LengthUnitLabels
                cbLower.Items.Add(sItem)
            Next
            cbLower.Visibility = Visibility.Visible
            cbLower.SelectedIndex = 0
            _GeomType = GeomType.Line
            _MeasureDrawObject.DrawMode = DrawMode.Polyline
            _MeasureDrawObject.IsEnabled = True
        Else
            lblUpperText.Text = "Perimeter: "
            cbUpper.Items.Clear()
            For Each sItem As String In _LengthUnitLabels
                cbUpper.Items.Add(sItem)
            Next
            cbUpper.Visibility = Visibility.Visible
            cbUpper.SelectedIndex = 0
            lblLowerText.Text = "Area: "
            cbLower.Items.Clear()
            For Each sItem As String In _AreaUnitLabels
                cbLower.Items.Add(sItem)
            Next
            cbLower.Visibility = Visibility.Visible
            cbLower.SelectedIndex = 0
            _GeomType = GeomType.Poly
            _MeasureDrawObject.DrawMode = DrawMode.Polygon
            _MeasureDrawObject.IsEnabled = True
        End If
        _Parent.SetCursor(MouseCursorType.Measure)
    End Sub

    Private Sub AddCurrentPoint()
        Dim pt As New MapPoint(_ptCurrent.X, _ptCurrent.Y)
        pt.SpatialReference = _Map.SpatialReference
        _MeasurePoints.Add(pt)
    End Sub

    Private Sub UpdateGeometry()

        Dim iCount As Integer
        Dim dLen, dSegLen, dArea As Double

        iCount = _MeasurePoints.Count()
        If iCount < 2 Then
            Exit Sub
        End If
        If _GeomType = GeomType.Line Then
            If _bGeodesic Then
                dLen = GeoLength(False)
                dSegLen = GeoDist(_MeasurePoints(iCount - 2), _MeasurePoints(iCount - 1))
            Else
                dLen = PlanarLength(False)
                dSegLen = Dist(_MeasurePoints(iCount - 2), _MeasurePoints(iCount - 1))
            End If
            _dUpperResult = dSegLen * _CF
            _dLowerResult = dLen * _CF
            UpdateValues()
            Exit Sub
        End If
        If iCount < 3 Then
            If _bGeodesic Then
                dLen = GeoLength(False)
            Else
                dLen = PlanarLength(False)
            End If
            _dUpperResult = dLen * _CF
            _dLowerResult = 0
            UpdateValues()
            Exit Sub
        End If
        If _bGeodesic Then
            dLen = GeoLength(True)
            dArea = GeoArea()
        Else
            dLen = PlanarLength(True)
            dArea = PlanarArea()
        End If
        _dUpperResult = dLen * _CF
        _dLowerResult = dArea * _CF * _CF
        UpdateValues()

    End Sub

    Private Sub UpdateValues()
        If _dUpperResult = 0 And _dLowerResult = 0 Then
            Exit Sub
        End If
        If _GeomType = GeomType.Line Then
            lblUpperText.Text = "Segment: " & ConvertLength(_dUpperResult)
            lblLowerText.Text = "Total: " & ConvertLength(_dLowerResult)
        Else
            lblUpperText.Text = "Perimeter: " & ConvertLength(_dUpperResult)
            lblLowerText.Text = "Area: " & ConvertArea(_dLowerResult)
        End If
    End Sub

    Private Function ConvertLength(ByVal dMeters As Double) As String
        Dim i As Integer
        If _btnCurrentTool Is btnLine Then
            i = cbLower.SelectedIndex
        Else
            i = cbUpper.SelectedIndex
        End If
        Dim dVal As Double = 0
        Select Case i
            Case 0 ' Miles
                dVal = dMeters / 0.3048 / 5280
            Case 1 ' Feet
                dVal = dMeters / 0.3048
            Case 2 ' Meters
                dVal = dMeters
            Case 3 ' Kilometers
                dVal = dMeters / 1000
        End Select
        Return dVal.ToString("0.##")
    End Function

    Private Function ConvertArea(ByVal dSqMeters As Double) As String
        Dim i As Integer = cbLower.SelectedIndex
        Dim dVal As Double = 0
        Select Case i
            Case 0 ' Sq. Miles
                dVal = dSqMeters / 0.09290304 / 27878400
            Case 1 ' Acres
                dVal = dSqMeters / 0.09290304 / 43560
            Case 2 ' Sq. Feet
                dVal = dSqMeters / 0.09290304
            Case 3 ' Sq. Meters
                dVal = dSqMeters
            Case 4 ' Sq. Kilometers
                dVal = dSqMeters / 1000000
        End Select
        Return dVal.ToString("0.####")
    End Function

    Private Function PlanarLength(ByVal bClose As Boolean) As Double
        Dim iCount, i As Integer
        Dim dLen As Double = 0
        iCount = _MeasurePoints.Count
        If iCount < 2 Then
            Return dLen
        End If
        For i = 0 To iCount - 2
            dLen += Dist(_MeasurePoints(i), _MeasurePoints(i + 1))
        Next
        If bClose Then
            dLen += Dist(_MeasurePoints(iCount - 1), _MeasurePoints(0))
        End If
        Return dLen
    End Function

    Private Function PlanarArea() As Double
        Dim dArea As Double = 0
        Dim iCount, i, j As Integer
        iCount = _MeasurePoints.Count
        If iCount < 3 Then
            Return 0
        End If
        For i = 0 To iCount - 1
            If i = iCount - 1 Then
                j = 0
            Else
                j = i + 1
            End If
            dArea += Determinant(_MeasurePoints(i), _MeasurePoints(j))
        Next
        Return Math.Abs(dArea / 2)
    End Function

    Private Function Dist(ByVal pt1 As MapPoint, ByVal pt2 As MapPoint) As Double
        Dim d1, d2 As Double
        d1 = pt2.X - pt1.X
        d2 = pt2.Y - pt1.Y
        Return Math.Sqrt((d1 * d1) + (d2 * d2))
    End Function

    Private Function Determinant(ByVal pt1 As MapPoint, ByVal pt2 As MapPoint) As Double
        Return (pt1.X * pt2.Y) - (pt2.X * pt1.Y)
    End Function
#End Region

#Region "Geodesic measurement functions"
    '**** From GRASS source code *****

    ' WGS-84 ellipsoid params
    Private Const _a As Double = 6378137
    Private Const _b As Double = 6356752.314

    ' Geodesic distance

    Private Function GeoLength(ByVal bClose As Boolean) As Double

        Dim iCount, i As Integer
        Dim dLen As Double = 0

        iCount = _MeasurePoints.Count
        If iCount < 2 Then
            Return dLen
        End If
        For i = 0 To iCount - 2
            dLen += GeoDist(_MeasurePoints(i), _MeasurePoints(i + 1))
        Next
        If bClose Then
            dLen += GeoDist(_MeasurePoints(iCount - 1), _MeasurePoints(0))
        End If
        Return dLen

    End Function

    Private Function GeoDist(ByVal pt1 As MapPoint, ByVal pt2 As MapPoint) As Double

        Dim lon1 As Double = pt1.X * Math.PI / 180
        Dim lat1 As Double = pt1.Y * Math.PI / 180
        Dim lon2 As Double = pt2.X * Math.PI / 180
        Dim lat2 As Double = pt2.Y * Math.PI / 180
        Dim e2 As Double = (_a * _a - _b * _b) / (_a * _a)

        Dim boa, f, ff64, al, t1, t2, t3, t4, t1r, t2r As Double
        Dim stm, ctm, sdtm, cdtm, tm, dtm As Double

        al = _a
        boa = Math.Sqrt(1 - e2)
        f = 1 - boa
        ff64 = f * f / 64

        t1r = Math.Atan(boa * Math.Tan(lat1))
        t2r = Math.Atan(boa * Math.Tan(lat2))
        tm = (t1r + t2r) / 2
        dtm = (t2r - t1r) / 2
        stm = Math.Sin(tm)
        ctm = Math.Cos(tm)
        sdtm = Math.Sin(dtm)
        cdtm = Math.Cos(dtm)
        t1 = stm * cdtm
        t1 = t1 * t1 * 2
        t2 = sdtm * ctm
        t2 = t2 * t2 * 2
        t3 = sdtm * sdtm
        t4 = cdtm * cdtm - stm * stm

        Dim a, cd, d, e, q, sd, sdlmr, t, u, v, x, y As Double

        sdlmr = Math.Sin((lon2 - lon1) / 2)
        If sdlmr = 0 And t1r = t2r Then
            Return 0
        End If
        q = t3 + sdlmr * sdlmr * t4
        If q = 1 Then
            Return Math.PI * al
        End If
        cd = 1 - 2 * q
        sd = 2 * Math.Sqrt(q - q * q)
        If q <> 0 And cd = 1 Then
            t = 1
        ElseIf sd = 0 Then
            t = 1
        Else
            t = Math.Acos(cd) / sd
        End If

        u = t1 / (1 - q)
        v = t2 / q
        d = 4 * t * t
        x = u + v
        e = -2 * cd
        y = u - v
        a = -d * e

        Return (al * sd * (t - f / 4 * (t * x - y) + ff64 * _
                (x * (a + (t - (a + e) / 2) * x) + y * (-2 * d + e * y) + d * x * y)))

    End Function

    ' Geodesic area

    Private QA, QB, QC As Double
    Private QbarA, QbarB, QbarC, QbarD As Double
    Private AE, Qp, E As Double

    Private Sub InitGeoAreaParams()

        Dim e2, e4, e6 As Double

        e2 = (_a * _a - _b * _b) / (_a * _a)
        e4 = e2 * e2
        e6 = e4 * e2

        AE = _a * _a * (1 - e2)
        QA = (2 / 3) * e2
        QB = (3 / 5) * e4
        QC = (4 / 7) * e6
        QbarA = -1 - (2 / 3) * e2 - (3 / 5) * e4 - (4 / 7) * e6
        QbarB = (2 / 9) * e2 + (2 / 5) * e4 + (4 / 7) * e6
        QbarC = -(3 / 25) * e4 - (12 / 35) * e6
        QbarD = (4 / 49) * e6

        Qp = Q(Math.PI / 2)
        E = 4 * Math.PI * Qp * AE
        If E < 0 Then
            E = -E
        End If

    End Sub

    Private Function GeoArea() As Double

        Dim iCount, i As Integer
        Dim x1, y1, x2, y2, dx, dy, Qbar1, Qbar2, area As Double

        iCount = _MeasurePoints.Count
        If iCount < 3 Then
            Return 0
        End If
        x2 = _MeasurePoints(iCount - 1).X * Math.PI / 180
        y2 = _MeasurePoints(iCount - 1).Y * Math.PI / 180
        Qbar2 = Qbar(y2)
        area = 0
        For i = 0 To iCount - 1
            x1 = x2
            y1 = y2
            Qbar1 = Qbar2
            x2 = _MeasurePoints(i).X * Math.PI / 180
            y2 = _MeasurePoints(i).Y * Math.PI / 180
            Qbar2 = Qbar(y2)
            If x1 > x2 Then
                While x1 - x2 > Math.PI
                    x2 += Math.PI + Math.PI
                End While
            ElseIf x2 > x1 Then
                While x2 - x1 > Math.PI
                    x1 += Math.PI + Math.PI
                End While
            End If
            dx = x2 - x1
            area += dx * (Qp - Q(y2))
            dy = y2 - y1
            If dy <> 0 Then
                area += dx * Q(y2) - (dx / dy) * (Qbar2 - Qbar1)
            End If
        Next
        area *= AE
        If area < 0 Then
            area = -area
        End If
        If area > E Then
            area = E
        End If
        If area > E / 2 Then
            area = E - area
        End If
        Return area

    End Function

    Private Function Q(ByVal x As Double) As Double
        Dim sinx, sinx2 As Double
        sinx = Math.Sin(x)
        sinx2 = sinx * sinx
        Return sinx * (1 + sinx2 * (QA + sinx2 * (QB + sinx2 * QC)))
    End Function

    Private Function Qbar(ByVal x As Double) As Double
        Dim cosx, cosx2 As Double
        cosx = Math.Cos(x)
        cosx2 = cosx * cosx
        Return cosx * (QbarA + cosx2 * (QbarB + cosx2 * (QbarC + cosx2 * QbarD)))
    End Function

#End Region

End Class
